#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['hotpot_km', 'hotpot_km.tests']

package_data = \
{'': ['*']}

install_requires = \
['async_generator', 'jupyter_client', 'nbformat', 'nest_asyncio', 'traitlets']

extras_require = \
{'mapping': ['jupyter_server'],
 'test': ['pytest', 'pytest-cov', 'jupyter_server', 'jupyter_client[test]']}

setup(name='hotpot_km',
      version='0.2.1',
      description='Hotpot - Jupyter kernel manager helpers',
      author='Vidar Tonaas Fauske',
      author_email='vidartf@gmail.com',
      url='https://github.com/voila-dashboards/hotpot_km',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='~=3.7',
     )
